/*


X3D Grouping Component

*/


/****************************************************************************
    This file is part of the FreeWRL/FreeX3D Distribution.

    Copyright 2009 CRC Canada. (http://www.crc.gc.ca)

    FreeWRL/FreeX3D is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeWRL/FreeX3D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeWRL/FreeX3D.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************************/



#include <config.h>
#include <system.h>
#include <display.h>
#include <internal.h>

#include <libFreeWRL.h>

#include "../vrml_parser/Structs.h"
#include "../vrml_parser/CRoutes.h"
#include "../main/headers.h"

#include "../opengl/OpenGL_Utils.h"
#include "../opengl/Frustum.h"
#include "../opengl/Material.h"

#include "LinearAlgebra.h"
#include "Children.h"
#include "../scenegraph/RenderFuncs.h"

typedef struct extent_type {
float e[6];
} extent_t;
typedef struct mat4x4_type {
	double mat[16];
} mat4x4_t;
typedef struct pComponent_Grouping{
	Stack *group_visible_stack;
	Stack *group_extent_stack;
	Stack *transform_local_stack;
}* ppComponent_Grouping;
void *Component_Grouping_constructor(){
	void *v = MALLOCV(sizeof(struct pComponent_Grouping));
	memset(v,0,sizeof(struct pComponent_Grouping));
	return v;
}
void Component_Grouping_init(struct tComponent_Grouping *t){
	//public

	//private
	t->prv = Component_Grouping_constructor();
	{
		ppComponent_Grouping p = (ppComponent_Grouping)t->prv;
		p->group_visible_stack = newStack(int);
		stack_push(int,p->group_visible_stack,TRUE); //need something/default to && with at top of stack
		p->group_extent_stack = newStack(extent_t);
		p->transform_local_stack = newStack(mat4x4_t);
	}
}
void Component_Grouping_clear(struct tComponent_Grouping *t){
	ppComponent_Grouping p = (ppComponent_Grouping)t->prv;
	deleteVector(struct X3D_Node*,p->transform_local_stack);
	deleteVector(struct X3D_Node*,p->group_extent_stack);
	deleteVector(struct X3D_Node*,p->group_visible_stack);
}



// v4 visibility functions, push & pop (to be) called from all X3DGroupingNode child_ functions
void push_group_visible(int visible){
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	stack_push(int,p->group_visible_stack,visible);
}
void pop_group_visible(){
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	stack_pop(int,p->group_visible_stack);
}
int peek_group_visible(){
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	return stack_top(int,p->group_visible_stack);
}

// v4 bbox functions, push & pop (to be) called from all X3DGroupingNode child_ functions
// children - geometry - will union their extent with the stack top owned by group with bbox (or scene root)
void push_group_extent(float *e6){
	extent_t e;
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	memcpy(&e.e,e6,6*sizeof(float));
	testVector(extent_t, p->group_extent_stack);
	stack_push(extent_t,p->group_extent_stack,e);
	testVector(extent_t, p->group_extent_stack);
}
void push_group_extent_default(){
	float extent6[6];
	extent6f_clear(extent6);
	push_group_extent(extent6);
}
void pop_group_extent(){
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	stack_pop(extent_t,p->group_extent_stack);
}
float * peek_group_extent(){
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	//extent_t* ee = (extent_t*)p->group_extent_stack->data;
	//int index = p->group_extent_stack->n - 1;
	//printf("ee[%d]=%p ", index, &ee[index]);
	//extent_t* en = stack_top_ptr(extent_t, p->group_extent_stack);
	//float* ef = en->e;
	//printf("stack_top_ptr= %p en->e %p \n", en, ef);
	return (stack_top_ptr(extent_t,p->group_extent_stack))->e;
}
void union_group_extent(float *e6){
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	float *etop = peek_group_extent();
	testVector(extent_t, p->group_extent_stack);
	extent6f_union_extent6f(etop,e6);
	testVector(extent_t, p->group_extent_stack);
}
void draw_bbox(float *center, float *size);


// v4 bbox functions, push & pop (to be) called from all X3DGroupingNode child_ functions
// children - geometry - will union their extent with the stack top owned by group with bbox (or scene root)
void push_transform_local(double *mat){
	mat4x4_t m4;
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	memcpy(&m4.mat,mat,16*sizeof(double));
	stack_push(mat4x4_t,p->transform_local_stack,m4);
}
void push_transform_local_identity(){
	double m4[16];
	matidentity4d(m4);
	push_transform_local(m4);
}
void pop_transform_local(){
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	stack_pop(mat4x4_t,p->transform_local_stack);
}
double * peek_transform_local(){
	ppComponent_Grouping p = (ppComponent_Grouping)gglobal()->Component_Grouping.prv;
	return (stack_top_ptr(mat4x4_t,p->transform_local_stack))->mat;
}
void reset_transform_local(double *mat){
	//no push or pop here, assume already pushed
	double *mattop = peek_transform_local();
	memcpy(mattop,mat,16*sizeof(double));
}
void multiply_transform_local(double *mat){
	//no push or pop here, assume already pushed
	double matboth[16];
	double *mattop = peek_transform_local();
	matmultiplyAFFINE(matboth,mat,mattop);
	memcpy(mattop,matboth,16*sizeof(double));
}
void compile_Transform (struct X3D_Transform *node) { 
	INITIALIZE_EXTENT;

	/* printf ("changed Transform for node %u\n",node); */
	node->__do_center = verify_translate ((GLfloat *)node->center.c);
	node->__do_trans = verify_translate ((GLfloat *)node->translation.c);
	node->__do_scale = verify_scale ((GLfloat *)node->scale.c);
	node->__do_rotation = verify_rotate ((GLfloat *)node->rotation.c);
	node->__do_scaleO = verify_rotate ((GLfloat *)node->scaleOrientation.c);

	node->__do_anything = (node->__do_center ||
			node->__do_trans ||
			node->__do_scale ||
			node->__do_rotation ||
			node->__do_scaleO);

	REINITIALIZE_SORTED_NODES_FIELD(node->children,node->_sortedChildren);
	MARK_NODE_COMPILED
}


/* we compile the Group so that children are not continuously sorted */
void compile_Group(struct X3D_Group *node) {
	REINITIALIZE_SORTED_NODES_FIELD(node->children,node->_sortedChildren);
	/*
	{
		int i;
		ConsoleMessage ("compile_Group, rootNode is %p",rootNode());
		for (i=0; i<node->children.n; i++) ConsoleMessage ("compile_Group %p, c %d is %p",node,i,node->children.p[i]);
		for (i=0; i<node->_sortedChildren.n; i++) ConsoleMessage ("compile_Group %p, sc %d is %p",node,i,node->_sortedChildren.p[i]);
	}
	*/
	MARK_NODE_COMPILED
}

/* prep_Group - we need this so that distance (and, thus, distance sorting) works for Groups */
void prep_Group (struct X3D_Group *node) {
	// printf ("prep_Group start dist %f\n",node->_dist);
	COMPILE_IF_REQUIRED

/* 
printf ("prepGroup %p (root %p), flags %x children %d ",node,rootNode,node->_renderFlags,node->children.n);
if ((node->_renderFlags & VF_Viewpoint) == VF_Viewpoint) printf ("VF_Viewpoint ");
if ((node->_renderFlags & VF_Geom) == VF_Geom) printf ("VF_Geom ");
if ((node->_renderFlags & VF_localLight) == VF_localLight) printf ("VF_localLight ");
if ((node->_renderFlags & VF_Sensitive) == VF_Sensitive) printf ("VF_Sensitive ");
if ((node->_renderFlags & VF_Blend) == VF_Blend) printf ("VF_Blend ");
if ((node->_renderFlags & VF_Proximity) == VF_Proximity) printf ("VF_Proximity ");
if ((node->_renderFlags & VF_Collision) == VF_Collision) printf ("VF_Collision ");
if ((node->_renderFlags & VF_globalLight) == VF_globalLight) printf ("VF_globalLight ");
if ((node->_renderFlags & VF_hasVisibleChildren) == VF_hasVisibleChildren) printf ("VF_hasVisibleChildren ");
if ((node->_renderFlags & VF_shouldSortChildren) == VF_shouldSortChildren) printf ("VF_shouldSortChildren ");*/
/*if ((node->_renderFlags & VF_inPickableGroup) == VF_inPickableGroup) printf ("VF_inPickableGroup "); */
/* printf ("\n"); */

}

/* do transforms, calculate the distance */
void prep_Transform (struct X3D_Transform *node) {
	COMPILE_IF_REQUIRED

	/* rendering the viewpoint means doing the inverse transformations in reverse order (while poping stack),
		* so we do nothing here in that case -ncoder */

	/* printf ("prep_Transform, render_hier vp %d geom %d light %d sens %d blend %d prox %d col %d\n",
	render_vp,render_geom,render_light,render_sensitive,render_blend,render_proximity,render_collision); */

	/* do we have any geometry visible, and are we doing anything with geometry? */
	OCCLUSIONTEST

	if(!renderstate()->render_vp) {
		/* do we actually have any thing to rotate/translate/scale?? */
		push_transform_local_identity();
		if (node->__do_anything) {

			FW_GL_PUSH_MATRIX(); //this one will persist till fin_Transform pops it
			FW_GL_PUSH_MATRIX(); //this is to get us a separate 4x4 matrix just for the stuff here
			FW_GL_LOAD_IDENTITY(); // .. wehich we will save for child_Transform to propagate its bbox up to its extent

			/* TRANSLATION */
			if (node->__do_trans)
				FW_GL_TRANSLATE_F(node->translation.c[0],node->translation.c[1],node->translation.c[2]);

			/* CENTER */
			if (node->__do_center)
				FW_GL_TRANSLATE_F(node->center.c[0],node->center.c[1],node->center.c[2]);

			/* ROTATION */
			if (node->__do_rotation) {
				FW_GL_ROTATE_RADIANS(node->rotation.c[3], node->rotation.c[0],node->rotation.c[1],node->rotation.c[2]);
			}

			/* SCALEORIENTATION */
			if (node->__do_scaleO) {
				FW_GL_ROTATE_RADIANS(node->scaleOrientation.c[3], node->scaleOrientation.c[0], node->scaleOrientation.c[1],node->scaleOrientation.c[2]);
			}


			/* SCALE */
			if (node->__do_scale)
				FW_GL_SCALE_F(node->scale.c[0],node->scale.c[1],node->scale.c[2]);

			/* REVERSE SCALE ORIENTATION */
			if (node->__do_scaleO)
				FW_GL_ROTATE_RADIANS(-node->scaleOrientation.c[3], node->scaleOrientation.c[0], node->scaleOrientation.c[1],node->scaleOrientation.c[2]);

			/* REVERSE CENTER */
			if (node->__do_center)
				FW_GL_TRANSLATE_F(-node->center.c[0],-node->center.c[1],-node->center.c[2]);
			{
				double mat[16];

				FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX,mat); //we got our local transform saved
				FW_GL_POP_MATRIX();
				FW_GL_TRANSFORM_D(mat); //now apply the above to prep for child_Tranform
				reset_transform_local(mat);
			}
		} 
	}
}


void fin_Transform (struct X3D_Transform *node) {
	OCCLUSIONTEST

	if(!renderstate()->render_vp) {
		pop_transform_local();
		if (node->__do_anything) {
			FW_GL_POP_MATRIX();
		}
	} else {
		/*Rendering the viewpoint only means finding it, and calculating the reverse WorldView matrix.*/
		if((node->_renderFlags & VF_Viewpoint) == VF_Viewpoint) 
		{
			FW_GL_TRANSLATE_F(((node->center).c[0]),((node->center).c[1]),((node->center).c[2])
			);
			FW_GL_ROTATE_RADIANS(((node->scaleOrientation).c[3]),((node->scaleOrientation).c[0]),((node->scaleOrientation).c[1]),((node->scaleOrientation).c[2])
			);
			FW_GL_SCALE_F((float)1.0/(((node->scale).c[0])),(float)1.0/(((node->scale).c[1])),(float)1.0/(((node->scale).c[2]))
			);
			FW_GL_ROTATE_RADIANS(-(((node->scaleOrientation).c[3])),((node->scaleOrientation).c[0]),((node->scaleOrientation).c[1]),((node->scaleOrientation).c[2])
			);
			FW_GL_ROTATE_RADIANS(-(((node->rotation).c[3])),((node->rotation).c[0]),((node->rotation).c[1]),((node->rotation).c[2])
			);
			FW_GL_TRANSLATE_F(-(((node->center).c[0])),-(((node->center).c[1])),-(((node->center).c[2]))
			);
			FW_GL_TRANSLATE_F(-(((node->translation).c[0])),-(((node->translation).c[1])),-(((node->translation).c[2]))
			);
		}
	}
} 

void child_Switch (struct X3D_Switch *node) {
	/* exceedingly simple - render only one child */
	struct X3D_Node **pp;
	int n;
	int wc = node->whichChoice;

	/* is this VRML, or X3D?? */
	n = 0;
	pp = NULL;
	//if(node->children.n){
	//if (node->__isX3D) {
	int spec = X3D_PROTO(node->_executionContext)->__specversion;
	if (spec < 300 || node->choice.n){
		pp = node->choice.p;
		n = node->choice.n;
	} else {
		pp = node->children.p;
		n = node->children.n;
		//} else if(node->choice.n){
	}
	if(n && pp){
		if(wc >= 0 && wc < n){
			void * p = pp[wc];

			prep_BBox((struct BBoxFields*)&node->bboxCenter);

			render_node(p);
		
			fin_BBox((struct X3D_Node*)node,(struct BBoxFields*)&node->bboxCenter,FALSE);
		}
	}
	//if (node->__isX3D ||  (node->children).n) {
	//	if(wc >= 0 && wc < (node->children).n) {
	//		void *p = ((node->children).p[wc]);
	//		render_node(p);
	//	}
	//} else {
	//	if(wc >= 0 && wc < ((node->choice).n)) {
	//		void *p = ((node->choice).p[wc]);
	//		render_node(p);
	//	}
	//}
}



void sib_prep_LocalFog(struct X3D_Node *parent, struct X3D_Node *sibAffector);
void sib_prep_Light(struct X3D_Node* parent, struct X3D_Node* sibAffector);
void sib_prep_ClipPlane(struct X3D_Node *parent, struct X3D_Node *sibAffector);
void sib_prep_Effect(struct X3D_Node *parent, struct X3D_Node *sibAffector);
void sib_prep_TextureProjector(struct X3D_Node *parent, struct X3D_Node *sibAffector);

void sib_prep(struct X3D_Node *parent, struct X3D_Node *sibAffector){
	switch(sibAffector->_nodeType){
		case NODE_DirectionalLight:
		case NODE_SpotLight:
		case NODE_PointLight:
			sib_prep_Light(parent, sibAffector);
			break;
		case NODE_LocalFog:
			sib_prep_LocalFog(parent,sibAffector); break;
		case NODE_ClipPlane:
			sib_prep_ClipPlane(parent,sibAffector); break;
		case NODE_Effect: 
			sib_prep_Effect(parent,sibAffector); break;
		case NODE_TextureProjector:
		case NODE_TextureProjectorParallel: 
		case NODE_TextureProjectorPoint:
			sib_prep_TextureProjector(parent,sibAffector); break;
		default:
			break;
	}
}

void sib_fin_LocalFog(struct X3D_Node *parent, struct X3D_Node *sibAffector);
void sib_fin_Light(struct X3D_Node* parent, struct X3D_Node* sibAffector);

void sib_fin_ClipPlane(struct X3D_Node *parent, struct X3D_Node *sibAffector);
void sib_fin_Effect(struct X3D_Node *parent, struct X3D_Node *sibAffector);
void sib_fin_TextureProjector(struct X3D_Node *parent, struct X3D_Node *sibAffector);

void sib_fin(struct X3D_Node *parent, struct X3D_Node *sibAffector){
	switch(sibAffector->_nodeType){
		case NODE_DirectionalLight:
		case NODE_SpotLight:
		case NODE_PointLight:
			sib_fin_Light(parent, sibAffector);
			break;
		case NODE_LocalFog:
			sib_fin_LocalFog(parent,sibAffector); break;
		case NODE_ClipPlane:
			sib_fin_ClipPlane(parent,sibAffector); break;
		case NODE_Effect:
			sib_fin_Effect(parent,sibAffector); break;
		case NODE_TextureProjector:
		case NODE_TextureProjectorParallel: 
		case NODE_TextureProjectorPoint:
			sib_fin_TextureProjector(parent,sibAffector); break;

		default:
			break;
	}
}
void prep_sibAffectors(struct X3D_Node *parent, struct Multi_Node* affectors){
	if(affectors->n){
		int j;
		for(j=0;j<affectors->n;j++){
			struct X3D_Node *sa = affectors->p[j];
			sib_prep(parent,sa);
		}
	}
}
void fin_sibAffectors(struct X3D_Node *parent, struct Multi_Node* affectors){
	if(affectors->n){
		int j,jj;
		for(jj=0;jj<affectors->n;jj++){
			//we go backwards so any sib_fin popping is in reverse order to any sib_prep pushing,
			// in case multiple push to same stack, as with multiple Effects
			struct X3D_Node *sa;
			j = affectors->n - jj - 1; 
			sa = affectors->p[j];
			sib_fin(parent,sa);
		}
	}
}

//WARNING all nodes that use prep_childrenBBox /fin_ musht have bbox fields in same order
// (this is a way to avoid macros, and re-casting lookups)
//struct BBoxFields {
//	struct SFVec3f bboxCenter;
//	struct SFVec3f bboxSize;
//	int visible;
//	int displayBBox;
//
//};
void prep_BBox(struct BBoxFields *bfields){
	ttrenderstate rs = renderstate(); //just want pure geom or geom+blend
	if (rs->render_vp || rs->render_other || rs->render_sensitive || rs->render_depth || rs->render_light || rs->render_cube || rs->render_collision) return;
	push_group_extent_default();
	push_group_visible( bfields->visible && peek_group_visible());
}

//#define VERBOSE
void fin_BBox(struct X3D_Node *node, struct BBoxFields *bfields, int transtype){
	ttrenderstate rs = renderstate(); //just want pure geom or geom+blend
	if (rs->render_vp || rs->render_other || rs->render_sensitive || rs->render_depth || rs->render_light || rs->render_cube || rs->render_collision) 
		return;

	#ifdef VERBOSE
	printf ("\nstart fin_BBox\n");
	printf ("... node %p type %s\n",node, stringNodeType(node->_nodeType));
	printf ("... center %4.3f %4.3f %4.3f, size %4.3f %4.3f %4.3f\n",
		bfields->bboxCenter.c[0],
		bfields->bboxCenter.c[1],
		bfields->bboxCenter.c[2],
		bfields->bboxSize.c[0],
		bfields->bboxSize.c[1],
		bfields->bboxSize.c[2]);
		
	#endif //VERBOSE

	pop_group_visible();
	//bbox - in child-space - gets transformed/propagated to Transform parent space and set as Transform._extent
	extent6f2bbox(peek_group_extent(),bfields->bboxCenter.c,bfields->bboxSize.c);

	#ifdef VERBOSE
	printf ("step1 fin_BBox\n");
	printf ("... node %p type %s\n",node, stringNodeType(node->_nodeType));
	printf ("... center %4.3f %4.3f %4.3f, size %4.3f %4.3f %4.3f\n",
		bfields->bboxCenter.c[0],
		bfields->bboxCenter.c[1],
		bfields->bboxCenter.c[2],
		bfields->bboxSize.c[0],
		bfields->bboxSize.c[1],
		bfields->bboxSize.c[2]);
	#endif //VERBOSE

	if(renderstate()->render_geom && (bfields->bboxDisplay || fwl_getDrawBoundingBoxes() )) {
		draw_bbox(bfields->bboxCenter.c,bfields->bboxSize.c);
	}
	//propagate bbox up one level
	if(transtype){
		//for transform type nodes, we capture a matrix in prep_Transform (and pop it in fin_Transform)
		//so we can transform the bbox up one level into parent space - so called 'propagating' extent
		extent6f_mattransform4d(node->_extent,peek_group_extent(),peek_transform_local());
	}else{
		//non-transforming grouping nodes - just copy bbox of children into parent space
		extent6f_copy(node->_extent,peek_group_extent());
	}


	#ifdef VERBOSE
	printf ("step2 fin_BBox\n");
	printf ("... node %p type %s\n",node, stringNodeType(node->_nodeType));
	printf ("... center %4.3f %4.3f %4.3f, size %4.3f %4.3f %4.3f\n",
		bfields->bboxCenter.c[0],
		bfields->bboxCenter.c[1],
		bfields->bboxCenter.c[2],
		bfields->bboxSize.c[0],
		bfields->bboxSize.c[1],
		bfields->bboxSize.c[2]);
	#endif //VERBOSE

	pop_group_extent(); // up where parents are
	union_group_extent(node->_extent); //

	record_ZBufferDistance(node,bfields);

	#ifdef VERBOSE
	printf ("node %p %s extent ",node,stringNodeType(node->_nodeType));
	int i;
	for (i=0; i<6; i++) printf ("%4.3f ",node->_extent[i]); printf ("\n");
	printf ("recording Z_BufferDistance at %s:%d\n",__FILE__,__LINE__);
	#endif //VERBOSE

	#ifdef VERBOSE
	printf ("fin fin_BBox nodeType %s\n",stringNodeType(node->_nodeType));
	#endif //VERBOSE
}
#undef VERBOSE


void child_StaticGroup (struct X3D_StaticGroup *node) {
	CHILDREN_COUNT
	//LOCAL_LIGHT_SAVE

	RETURN_FROM_CHILD_IF_NOT_FOR_ME

	/* did this change? */
	if NODE_NEEDS_COMPILING {
		REINITIALIZE_SORTED_NODES_FIELD(node->children,node->_sortedChildren);
		//ConsoleMessage ("StaticGroup changed");
		MARK_NODE_COMPILED;
	}

	/* do we have a local light for a child? */
	//LOCAL_LIGHT_CHILDREN(node->_sortedChildren);
	prep_sibAffectors((struct X3D_Node*)node,&node->__sibAffectors);

	/* now, just render the non-directionalLight children */
	prep_BBox((struct BBoxFields*)&node->bboxCenter);
	normalChildren(node->_sortedChildren);
	fin_BBox((struct X3D_Node*)node,(struct BBoxFields*)&node->bboxCenter,FALSE);

	//LOCAL_LIGHT_OFF
	fin_sibAffectors((struct X3D_Node*)node,&node->__sibAffectors);

}

void child_Group (struct X3D_Group *node) {
	// UNUSED int renderFirstProtoChildOnlyAsPerSpecs = 1;
	CHILDREN_COUNT
//	LOCAL_LIGHT_SAVE

	/*
printf ("chldGroup %p (root %p), flags %x children %d ",node,rootNode,node->_renderFlags,node->children.n);
if ((node->_renderFlags & VF_Viewpoint) == VF_Viewpoint) printf ("VF_Viewpoint ");
if ((node->_renderFlags & VF_Geom) == VF_Geom) printf ("VF_Geom ");
if ((node->_renderFlags & VF_localLight) == VF_localLight) printf ("VF_localLight ");
if ((node->_renderFlags & VF_Sensitive) == VF_Sensitive) printf ("VF_Sensitive ");
if ((node->_renderFlags & VF_Blend) == VF_Blend) printf ("VF_Blend ");
if ((node->_renderFlags & VF_Proximity) == VF_Proximity) printf ("VF_Proximity ");
if ((node->_renderFlags & VF_Collision) == VF_Collision) printf ("VF_Collision ");
if ((node->_renderFlags & VF_globalLight) == VF_globalLight) printf ("VF_globalLight ");
if ((node->_renderFlags & VF_hasVisibleChildren) == VF_hasVisibleChildren) printf ("VF_hasVisibleChildren ");
if ((node->_renderFlags & VF_shouldSortChildren) == VF_shouldSortChildren) printf ("VF_shouldSortChildren ");
printf ("\n");
*/

	RETURN_FROM_CHILD_IF_NOT_FOR_ME

	if(1){
		//stereoscopic experiments
		ttrenderstate rs = renderstate();
		if (rs->render_geom) { //== VF_Geom) {
			if (node->_renderFlags & VF_HideLeft && (viewer_iside() == 0) )  { 
					return; 
			} 
			if (node->_renderFlags & VF_HideRight && (viewer_iside() == 1) )  { 
					return; 
			} 
		} 
	}
	prep_sibAffectors((struct X3D_Node*)node,&node->__sibAffectors);


#ifdef VERBOSE
	 {
		int x;
		struct X3D_Node *xx;

printf ("child_Group,  children.n %d sortedChildren.n %d\n",node->children.n, node->_sortedChildren.n);

		printf ("child_Group, this %p rf %x isProto %d\n",node,node->_renderFlags, node->FreeWRL__protoDef);
//        printf ("	..., render_hier vp %d geom %d light %d sens %d blend %d prox %d col %d\n",
//         render_vp,render_geom,render_light,render_sensitive,render_blend,render_proximity,render_collision); 
		for (x=0; x<nc; x++) {
			xx = X3D_NODE(node->_sortedChildren.p[x]);
			if (xx)
			printf ("	%d: ch %p type %s dist %f\n",x, node->_sortedChildren.p[x],stringNodeType(xx->_nodeType),xx->_dist);
			else printf ("     chiuld %d null\n",x);
		}
		for (x=0; x<nc; x++) {
			xx = X3D_NODE(node->_sortedChildren.p[x]);
			if (xx)
			printf ("	%d: sch %p type %s dist %f\n",x, node->_sortedChildren.p[x],stringNodeType(xx->_nodeType),xx->_dist);
			else printf ("     chiuld %d null\n",x);
		}
	}
#endif //VERBOSE

	/* do we have a DirectionalLight for a child? */
//	LOCAL_LIGHT_CHILDREN(node->_sortedChildren);

	/* printf ("chld_Group, for %u, protodef %d and FreeWRL_PROTOInterfaceNodes.n %d\n",
		node, node->FreeWRL__protoDef, node->FreeWRL_PROTOInterfaceNodes.n); */
	/* now, just render the non-directionalLight children */
	// UNUSED renderFirstProtoChildOnlyAsPerSpecs = 0; //flux/vivaty render all children
	prep_BBox((struct BBoxFields*)&node->bboxCenter);
	normalChildren(node->_sortedChildren);
	fin_BBox((struct X3D_Node*)node,(struct BBoxFields*)&node->bboxCenter,FALSE);

	#ifdef VERBOSE
	printf ("child_Group ");
	//WARNING all nodes that use prep_childrenBBox /fin_ musht have bbox fields in same order
	// (this is a way to avoid macros, and re-casting lookups)
	//struct BBoxFields {
	//      struct SFVec3f bboxCenter;
	//      struct SFVec3f bboxSize;
	//      int visible;
	//      int displayBBox;
	//
	//};
	printf ("bboxCenter %4.3f %4.3f %4.3f ",node->bboxCenter.c[0],node->bboxCenter.c[1],node->bboxCenter.c[2]);
	printf ("bboxSize %4.3f %4.3f %4.3f ",node->bboxSize.c[0],node->bboxSize.c[1],node->bboxSize.c[2]);
	printf ("visible %d displayBox %d\n",node->visible,node->bboxDisplay);
	#endif //VERBOSE



//	LOCAL_LIGHT_OFF
	
	fin_sibAffectors((struct X3D_Node*)node,&node->__sibAffectors);
}


void child_Transform (struct X3D_Transform *node) {
	// printf ("start child_Transform, dist %f\n",node->_dist);
	//LOCAL_LIGHT_SAVE
	CHILDREN_COUNT
	OCCLUSIONTEST

	RETURN_FROM_CHILD_IF_NOT_FOR_ME

	if(1){
		//stereoscopic experiments
		ttrenderstate rs = renderstate();
		if (rs->render_geom) { //== VF_Geom) {
			if (node->_renderFlags & VF_HideLeft && (viewer_iside() == 0) )  { 
					return; 
			} 
			if (node->_renderFlags & VF_HideRight && (viewer_iside() == 1) )  { 
					return; 
			} 
		} 
	}

	/* any children at all? */
	if (nc==0) return;

	//if(node->__sibAffectors.n)
	//	printf("have transform sibaffectors\n");
	prep_sibAffectors((struct X3D_Node*)node,&node->__sibAffectors);


	//profile_start("local_light_kids");
	/* do we have a local light for a child? */
//	LOCAL_LIGHT_CHILDREN(node->_sortedChildren);
	//profile_end("local_light_kids");
	/* now, just render the non-directionalLight children */

	/* printf ("Transform %d, flags %d, render_sensitive %d\n",
			node,node->_renderFlags,render_sensitive); */

	#ifdef CHILDVERBOSE
		printf ("transform - doing normalChildren\n");
	#endif

	prep_BBox((struct BBoxFields*)&node->bboxCenter);
	normalChildren(node->_sortedChildren);
	fin_BBox((struct X3D_Node*)node,(struct BBoxFields*)&node->bboxCenter,TRUE);

	#ifdef CHILDVERBOSE
		printf ("transform - done normalChildren\n");
	#endif

//	LOCAL_LIGHT_OFF
	fin_sibAffectors((struct X3D_Node*)node,&node->__sibAffectors);
	//printf ("fin child_Transform, dist %f\n",node->_dist);
}


/* prep_Proto - this is a ProtoInstance (not declare)  */
/*pack 4 flags into one int, using char */
char ciflag_get(int flags, int index){
	char *cflags = (char *)(&flags);
	return cflags[index];
}
int ciflag_set(int flags, char flag, int index ){
	char *cflags = (char *)(&flags);
	cflags[index] = flag;
	return flags;
}
void prep_Proto (struct X3D_Proto *node) {
	if(0)printf("in prep_proto\n");
	load_externProtoInstance(node);
	COMPILE_IF_REQUIRED
	//RECORD_DISTANCE
}
/* not sure why we would compile */
void compile_Proto(struct X3D_Proto *node) {
	unsigned char pflag;
	if(0)printf("in compile_proto\n");
	pflag = ciflag_get(node->__protoFlags,2);
	if(pflag == 2){
		//scene
		REINITIALIZE_SORTED_NODES_FIELD(node->__children,node->_sortedChildren);
	}
	MARK_NODE_COMPILED
}


//UNIT statement - applying unit scalefactor during rendering
int doLengthUnits();
int isUnitSpecVersionOK(int specversion);
static int unitoption_scalescene = TRUE; //FALSE is this what specs say, don't scale top scene? 
void prep_unitscale (struct X3D_Proto *ec) {
	//if(doLengthUnits() && isUnitSpecVersionOK(ec->__specversion)){
	if(doLengthUnits()){
		if(!renderstate()->render_vp) {
			struct X3D_Proto *parent;
			double factor = 1.0; //default factor applied if current context < v3.3
			double parentfactor = 1.0;  //default factor applied if parent context < v3.3 or no parent
			parent = X3D_PROTO(ec->_executionContext); //__parentProto); //not sure this is correct. Looking for parent context of Instance, not ProtoDefinition
			FW_GL_PUSH_MATRIX();
			// SCALE 
			if(isUnitSpecVersionOK(ec->__specversion)){
				factor = ec->__unitlengthfactor;
			}
			if(parent){
				if(isUnitSpecVersionOK(parent->__specversion))
					parentfactor = parent->__unitlengthfactor;
			}else {
				//there's no higher level context, which means we are in the top scene
				if(unitoption_scalescene){
					parentfactor = 1.0; //scale top scene length to SI base units [m] (not what specs say)
				}else{
					parentfactor = factor; 
					//top level scene gets scale of 1 ie if its in feet it stays feet, .3048/.3048=1 as per specs
					//this is so defaults like Sphere radius=1 will be 1 foot, no need to tinker with defaults
				}
			}
			//printf("( factor %lf / parentfactor= %lf  ", factor, parentfactor);
			factor = factor / parentfactor;
			//printf(" = %lf)\n",factor);
			FW_GL_SCALE_D(factor,factor,factor);
			//RECORD_DISTANCE
		}
	}
}


void fin_unitscale (struct X3D_Proto *ec) {

	//if(doLengthUnits() && isUnitSpecVersionOK(ec->__specversion)){
	if(doLengthUnits()){
		if(!renderstate()->render_vp) {
			FW_GL_POP_MATRIX();
		} 
		/*
		else {
			//Rendering the viewpoint only means finding it, and calculating the reverse WorldView matrix.
			if((ec->_renderFlags & VF_Viewpoint) == VF_Viewpoint) {
				struct X3D_Proto *parent;
				double factor = 1.0;
				double parentfactor = 1.0;
				parent = X3D_PROTO(ec->__parentProto); //not sure this is correct. Looking for parent context of Instance, not ProtoDefinition
				if(parent)
					parentfactor = parent->__unitlengthfactor;
				FW_GL_PUSH_MATRIX();
				// SCALE 
				factor = ec->__unitlengthfactor;
				factor = parentfactor / factor;

				FW_GL_SCALE_D(factor,factor,factor);
			}
		}
		*/
	}
} 


/* render the first node only unless scene (see component_networking.c child_inline for scene-similar*/
void child_Proto (struct X3D_Proto *node) {
	int nc;
	unsigned char sceneflag;
	int renderFirstProtoChildOnlyAsPerSpecs;
	//LOCAL_LIGHT_SAVE
	if(0)printf("in child_proto\n");
	//CHILDREN_COUNT
	nc = node->__children.n; //_sortedChildren.n;
/*
printf ("chldGroup %p (root %p), flags %x children %d ",node,rootNode,node->_renderFlags,node->children.n);
if ((node->_renderFlags & VF_Viewpoint) == VF_Viewpoint) printf ("VF_Viewpoint ");
if ((node->_renderFlags & VF_Geom) == VF_Geom) printf ("VF_Geom ");
if ((node->_renderFlags & VF_localLight) == VF_localLight) printf ("VF_localLight ");
if ((node->_renderFlags & VF_Sensitive) == VF_Sensitive) printf ("VF_Sensitive ");
if ((node->_renderFlags & VF_Blend) == VF_Blend) printf ("VF_Blend ");
if ((node->_renderFlags & VF_Proximity) == VF_Proximity) printf ("VF_Proximity ");
if ((node->_renderFlags & VF_Collision) == VF_Collision) printf ("VF_Collision ");
if ((node->_renderFlags & VF_globalLight) == VF_globalLight) printf ("VF_globalLight ");
if ((node->_renderFlags & VF_hasVisibleChildren) == VF_hasVisibleChildren) printf ("VF_hasVisibleChildren ");
if ((node->_renderFlags & VF_shouldSortChildren) == VF_shouldSortChildren) printf ("VF_shouldSortChildren ");

printf ("\n");
*/
	RETURN_FROM_CHILD_IF_NOT_FOR_ME
	//if(node->__loadstatus != LOAD_STABLE) return; #define LOAD_STABLE 10
	push_executionContext(X3D_NODE(node));
	prep_unitscale(node);

#ifdef VERBOSE
	 {
		int x;
		struct X3D_Node *xx;

printf ("child_Group,  children.n %d sortedChildren.n %d\n",node->children.n, node->_sortedChildren.n);

		printf ("child_Group, this %p rf %x isProto %d\n",node,node->_renderFlags, node->FreeWRL__protoDef);
//        printf ("	..., render_hier vp %d geom %d light %d sens %d blend %d prox %d col %d\n",
//         render_vp,render_geom,render_light,render_sensitive,render_blend,render_proximity,render_collision); 
		for (x=0; x<nc; x++) {
			xx = X3D_NODE(node->_sortedChildren.p[x]);
			if (xx)
			printf ("	%d: ch %p type %s dist %f\n",x, node->_sortedChildren.p[x],stringNodeType(xx->_nodeType),xx->_dist);
			else printf ("     chiuld %d null\n",x);
		}
		for (x=0; x<nc; x++) {
			xx = X3D_NODE(node->_sortedChildren.p[x]);
			if (xx)
			printf ("	%d: sch %p type %s dist %f\n",x, node->_sortedChildren.p[x],stringNodeType(xx->_nodeType),xx->_dist);
			else printf ("     chiuld %d null\n",x);
		}
	}
#endif //VERBOSE



	prep_sibAffectors((struct X3D_Node*)node,&node->__sibAffectors);
	prep_BBox((struct BBoxFields*)&node->bboxCenter);


	sceneflag = ciflag_get(node->__protoFlags,2);
	renderFirstProtoChildOnlyAsPerSpecs = TRUE;  //FALSE is like flux / vivaty
	//I don't think inline.children comes through here, just scene and protoInstance
	if(sceneflag == 2 ){ 
		normalChildren(node->_sortedChildren);
	}else{
		if(renderFirstProtoChildOnlyAsPerSpecs && (renderstate()->render_geom || renderstate()->render_blend)) {
			//H: its just when rendering drawable geometry that we take only the first node
			(node->__children).n = 1;
			normalChildren(node->__children);
			(node->__children).n = nc;
		} else {
			//H: else even for Protobodies, we may visit all rootnodes & descendants
			//  in case they need some updating on a non-draw scenegraph pass?
			normalChildren(node->__children);
		}
	}

	fin_BBox((struct X3D_Node*)node,(struct BBoxFields*)&node->bboxCenter,FALSE);

	fin_sibAffectors((struct X3D_Node*)node,&node->__sibAffectors);
	fin_unitscale(node);
	pop_executionContext();

}

